;
;   Semi.lisp Thomas MANGIN
;

(setq Symbole   nil)    ; List of the substitution symboles

(setq RuleIn    nil)    ; Rule to find
(setq RuleOut   nil)    ; Rule to apply

(setq Expression nil)   ; expression to simplify
(setq DebugLevel 0)     ; Information output level



;
; Global values manipulation
;----------------------------

(defun set_expression (list)
  (setq Expression list)
 )


(defun set_debug_level (level)
  (setq DebugLevel level)
 )



;
; Initialisation of the Symbole list
;------------------------------------

(defun set_symbole (list)
  (if list
    (progn
      (push (car list) Symbole)
      (set_symbole  (cdr list))
     )
    't
   )
 )



;
; Read the rules to apply 
;-------------------------

(defun set_rule (list)
  (if list
    (let ((rule (car list)))
      (progn
        (push (car  rule) RuleIn )
        (push (cadr rule) RuleOut)
        (set_rule     (cdr  list))
       )
     )
    't
   )
 )



;
; Is the symbole sustituable 
;-----------------------------

(defun is_symbole (element)
  (member element Symbole)
 )



;
; User Interface (input)
;------------------------

(defun get_symbole ()
  (print "Enter the Substitution symbols.")
  (set_symbole(read))
 )

(defun get_rule ()
  (print "Enter the rule.")
  (set_rule(read))
 )

(defun get_expression ()
  (print "Enter the expression")
  (set_expression (read))
 )


(defun get_debug_level ()
  (print "Enter debug level")
  (set_debug_level (get_number 2))
 )



;
; SemiUnification Engine
;------------------------


(defun semiunification_init ())  ; Read the user input
(defun semiunification_run ())   ; While a rule is applyed run main loop
(defun reduce ())                ; Test all rules
(defun transform ())             ; Cut the function in sub-list to be checked
(defun substitute ())            ; Substitute its rule if possible
(defun can_substitute ())        ; Test if substitution is possible

(defun intern_substitute ())     ; Internal functions due to language 
(defun intern_can_substitute ()) ; recursive structure


;
; Reading all the rule
;-----------------------

(defun semiunification_init ()
  (print "Reading rules")
  (get_symbole)
  (get_rule)
  (get_expression)
  (get_debug_level)
  (quote "Reading done")
 )

;
; Main function
;----------------

(defun semiunification_run ()
  (let ( (result (reduce RuleIn RuleOut Expression)) )

    ; It was impossible to reduce any more
    (if (equal Expression result)
      ; Return finl function
      result

      ; Continue SemiUnification 
      (progn
        (setq Expression result)
        (semiunification_run)
       )
     )
   )
 )



;
; Apply all the rules one per one
;----------------------------------

(defun reduce (rule_in rule_out function)

  ; Isolate the new rule
  (let (
         (in  (car rule_in ))
         (out (car rule_out))
        )

    ; If still have rule
    (if in

      ; try to apply the rule to function
      (let ((result (transform in out function) ))

        ; If debug report
        (if
          (and
            (equal DebugLevel 1)
            (not (equal function result))
           )
            (progn
            (print "En Applicant")
            (print in)
            (print "On obtient")
            (print result)
           )
          nil
         )

        ; Try next rule
        (reduce (cdr rule_in) (cdr rule_out) result)
       )
      function
     )
   )
 )



;
; Try to transform the list if not possible its car and cdr
;-----------------------------------------------------------

(defun transform (rule_in rule_out function)

  ; Is list and not empty
  (if
    (and
      function
      (listp function)
     )

    ; Try to substitute the rule
    (let ((substitued (substitute rule_in rule_out function)))

      (if substitued

        ; If succes return result
		substitued

        ;create result list
        (cons
		  (transform rule_in rule_out (car function))
		  (transform rule_in rule_out (cdr function))
		 )
       )
     )
    function
   )
 )

;
; Try to modify the function with the rules
;-------------------------------------------

(defun substitute (rule_in rule_out function)
  (if (can_substitute rule_in function)
    (intern_substitute rule_out)
    nil
   )
 )

(defun intern_substitute (new)
  (if (stack_empty)
    new
    (intern_substitute (subst (pop_rule) (pop_motif) new))
   )
 )

;
; Check if the function contain the matched rule
;------------------------------------------------

(defun can_substitute (rule function)
  (progn
    (stack_reset)
    (intern_can_substitute rule function)
   )
 )

(defun intern_can_substitute (rule function)
  (if (and rule function)
    (let (
           (start_rule     (car rule))
           (start_function (car function))
          )
      (if
        (or
          ; The substitution is possible
          (and
            start_function
            (is_symbole start_rule)
            (push_couple start_rule start_function)
           )

          ; Exact rule match
          (equal start_rule start_function)

          ; The car is a list checked and valid
          (and
            (not (atomp start_rule))
            (not (atomp start_function))
            (listp start_rule)
            (listp start_function)
            (intern_can_substitute (car rule) (car function))
           )
         )

        (intern_can_substitute (cdr rule) (cdr function))
        nil
      )
    )
    (equal rule function) ; return (Rule == Function) (can only be nil)
  )
)


